package de.unibi.airobots.bonsai.actuators;

import java.io.IOException;
import java.util.LinkedList;
import java.util.Vector;

import de.unibi.airobots.bonsai.core.Actuator;
import de.unibi.airobots.bonsai.util.helper.ArmStatus;
import de.unibi.airobots.btl.data.KatanaGripperData;
import de.unibi.airobots.btl.data.KatanaPoseData;
import de.unibi.airobots.btl.data.PoseData;

public interface PoseActuatorTobi extends Actuator {

    /**
     * Move a Joint of the arm.
     * 
     * @param joint
     * @param value
     * @throws IOException
     */
    void moveJoint(int joint, double value) throws IOException;

    /**
     * List all angles.
     * 
     * @return joints in a LinkList
     * @throws IOException
     */
    LinkedList<Double> listJoints() throws IOException;

    /**
     * Goto a single Position (fully described by P,N,O,A).
     * 
     * @param px
     * @param py
     * @param pz
     * @param nx
     * @param ny
     * @param nz
     * @param ax
     * @param ay
     * @param az
     * @param ox
     * @param oy
     * @param oz
     * @return <code>true</code> if the position is reachable,
     *         <code>false</code> if not
     * @throws IOException
     */
    boolean goTo(double px, double py, double pz, double nx, double ny,
            double nz, double ax, double ay, double az, double ox, double oy,
            double oz);

    /**
     * List all available Movements.
     * 
     * @return Movements in a LinkedList
     * @throws IOException
     */
    Vector<String> listMovements() throws IOException;

    /**
     * Set a special Movement.
     * 
     * @param name
     * @throws IOException
     */
    void setMovements(String name) throws IOException;

    /**
     * Open the Gripper of the arm.
     * 
     * @throws IOException
     */
    void openGripper() throws IOException;

    /**
     * Close the Gripper of the arm.
     */
    void closeGripper() throws IOException;

    /**
     * Open gripper when touching
     * 
     * @throws IOException
     */
    void openGripperWhenTouching(int waitSeconds) throws IOException;

    /**
     * Get the Gripper sensor data
     * 
     * @return sensor data of the gripper
     * @throws IOException
     */
    KatanaGripperData getGipperSensorData() throws IOException;

    /**
     * go to a position
     * 
     * @param position
     *            go to this position
     * @param linear
     *            linear movement or non-linear movement
     */
    boolean goTo(KatanaPoseData position, boolean linear) throws IOException;

    /**
     * execute grasping Strategy
     * 
     * @param targetPose
     *            position of the object
     * @throws IOException
     */
    ArmStatus executeStrategy(PoseData targetPose) throws IOException;

    /**
     * Closing Gripper until a define force is achieved
     * 
     * @throws IOException
     */
    void closeGripperByForce() throws IOException;

}
