package de.unibi.airobots.bonsai.xcf;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;

import net.sf.xcf.AsyncCallback;
import net.sf.xcf.MethodNotFoundException;
import net.sf.xcf.RemoteServer;
import net.sf.xcf.UserException;
import net.sf.xcf.XcfException;
import net.sf.xcf.event.Xcf1Signature;
import net.sf.xcf.event.system.SystemEventListener;
import net.sf.xcf.ice.XCF.Remote.ServerPrx;
import net.sf.xcf.transport.XOPData;

/**
 * A generic wrapper (decorator) for {@link RemoteServer}s that allows flexible
 * actions on calls to remote server methods. Therefore java's reflection
 * mechanism is used. Subclasses have to implement
 * {@link #reflectiveCallMethod(RemoteServerMethod, Object...)} in order to
 * provide a custom way of handling method calls and wrapping them in utility
 * code. To simplify the reflection work they should use
 * {@link #invokeWrappedServerMethod(RemoteServerMethod, Object...)} to really
 * do the method call and put their error handling code etc. around a call to
 * this method.
 * 
 * @author jwienke
 */
@SuppressWarnings("deprecation")
public abstract class RemoteServerWrapper extends RemoteServer {

    private Logger logger = Logger.getLogger(getClass());

    private static final String CALL_METHOD_NAME = "callMethod";

    /**
     * Wrapped methods of {@link RemoteServer} with their method names and
     * signatures.
     * 
     * @author jwienke
     */
    protected enum RemoteServerMethod {

        /**
         * Signature of {@link RemoteServer#activate()}.
         */
        ACTIVATE("activate"),

        /**
         * Signature of {@link RemoteServer#deactivate()}.
         */
        DEACTIVATE("deactivate"),

        /**
         * Signature of {@link RemoteServer#addListener(SystemEventListener)}.
         */
        ADD_LISTENER("addListener", SystemEventListener.class),

        /**
         * Signature of {@link RemoteServer#removeListener(SystemEventListener)}
         * .
         */
        REMOVE_LISTENER("removeListener", SystemEventListener.class),

        /**
         * Signature of {@link RemoteServer#callMethod(String, XOPData)}.
         */
        NORMAL_CALL_METHOD(CALL_METHOD_NAME, String.class, XOPData.class),

        /**
         * Signature of
         * {@link RemoteServer#callMethod(String, XOPData, AsyncCallback)}.
         */
        CALL_METHOD_CALLBACK(CALL_METHOD_NAME, String.class, XOPData.class,
                AsyncCallback.class),

        /**
         * Signature of
         * {@link RemoteServer#callMethod(String, XOPData, Xcf1Signature, AsyncCallback)}
         * .
         */
        CALL_METHOD_XCF1_CALLBACK(CALL_METHOD_NAME, String.class,
                XOPData.class, Xcf1Signature.class, AsyncCallback.class),

        /**
         * Signature of
         * {@link RemoteServer#callMethod(String, XOPData, Xcf1Signature)}.
         */
        CALL_METHOD_XCF1(CALL_METHOD_NAME, String.class, XOPData.class,
                Xcf1Signature.class),

        /**
         * Signature of {@link RemoteServer#getMethods()}.
         */
        GET_METHODS("getMethods"),

        /**
         * Signature of {@link RemoteServer#getServerName()}.
         */
        GET_SERVER_NAME("getServerName");

        private String methodName;
        private Class<?>[] signature;

        RemoteServerMethod(String methodName, Class<?>... signature) {
            this.methodName = methodName;
            this.signature = signature;
        }

        String getMethodName() {
            return this.methodName;
        }

        Class<?>[] getSignature() {
            return this.signature;
        }

        @Override
        public String toString() {
            StringBuilder signatureBuilder = new StringBuilder();
            for (Class<?> c : signature) {
                signatureBuilder.append(c.getSimpleName());
                signatureBuilder.append(", ");
            }
            if (signatureBuilder.length() > 2) {
                CharSequence sig =
                        signatureBuilder.subSequence(0, signatureBuilder
                                .length() - 2);
                return super.toString() + "[methodName = " + methodName
                        + ", signature = " + sig + "]";
            } else {
                return super.toString() + "[methodName = " + methodName
                        + ", signature = " + signatureBuilder.toString() + "]";
            }
        }

    }

    /**
     * Reference to the {@link RemoteServerReconnectableManager} that manages
     * this wrapper.
     */
    private RemoteServerReconnectableManager manager;

    private static final String REFLECTION_ERROR_MESSAGE =
            "Cannot access RemoteServer methods "
                    + "via reflection as expected. "
                    + "Either RemoteServer API has changed "
                    + "or this is a serious programming error.";

    /**
     * A map of method objects use to call the remote server. This map is needed
     * instead of direct method objects for each method to invoke because a
     * wrapped method can be called before the first wrapped remote server is
     * present. This in turn means that all remote server methods would be
     * <code>null</code> and cannot be passed to the reflective call method.
     * Therefore method name keys are used.
     */
    private Map<RemoteServerMethod, Method> methodsByUniqueName =
            new HashMap<RemoteServerMethod, Method>();

    /**
     * The wrapped {@link RemoteServer} instance. This instance may be replaced
     * during runtime in a separate thread. <code>null</code> means there is no
     * initial remote server set and you need to wait for a new one.
     * 
     * Always acquire {@link #remoteServerLock} before accessing this object;
     */
    private RemoteServer wrappedRemoteServer = null;

    /**
     * A condition variable and lock object used to wait for a new remote server
     * passed in externally if the current remote server is disconnected.
     */
    private Object remoteServerLock = new Object();

    /**
     * Constructor.
     * 
     * @param wrappedServer
     *            wrapped server, may be <code>null</code> if not yet present
     * @param manager
     *            manager managing this instance
     */
    protected RemoteServerWrapper(RemoteServer wrappedServer,
            RemoteServerReconnectableManager manager) {
        // XXX I know that this works, but that is bad internal knowledge
        super(null, null);

        logger.debug("Constructor called");

        if (wrappedServer != null) {
            resetRemoteServer(wrappedServer);
        }
        this.manager = manager;
    }

    /**
     * Hook method that must be implemented to provide a custom way to call
     * methods on the wrapped remote server with a custom error handling.
     * 
     * @param method
     *            the method that should be called as decided by the client.
     *            Simply pass this to
     *            {@link #invokeWrappedServerMethod(RemoteServerMethod, Object...)}
     * @param args
     *            arguments for the wrapped method. Simply pass this to
     *            {@link #invokeWrappedServerMethod(RemoteServerMethod, Object...)}
     * @return return the result of
     *         {@link #invokeWrappedServerMethod(RemoteServerMethod, Object...)}
     * @throws RuntimeException
     *             indicates a serious programming error and should not be
     *             caught
     * @throws UserException
     *             a valid user exception from the wrapped server
     * @throws MethodNotFoundException
     *             a valid user exception from the wrapped server
     */
    // CHECKSTYLE:OFF we really want to throw everything that is possible here
    // because XCF already does this.
    protected abstract Object reflectiveCallMethod(RemoteServerMethod method,
            Object... args) throws RuntimeException, UserException,
            MethodNotFoundException;

    // CHECKSTYLE:ON

    /**
     * Returns the method object to the remote server method with the given
     * specification.
     * 
     * @param method
     *            method to retrieve
     * @return method or <code>null</code> if not found
     */
    protected Method getMethod(RemoteServerMethod method) {
        return methodsByUniqueName.get(method);
    }

    /**
     * Invokes a method on the wrapped remote server.
     * 
     * @param method
     *            the method to invoke
     * @param args
     *            arguments to pass to the method
     * @return the methods return value or <code>null</code> for void methods
     * @throws InvocationTargetException
     *             an error occurred invoking the method that is not related to
     *             a reflection problem. This error (or its causing exception)
     *             should be handled by the subclasses.
     * @throws RuntimeException
     *             this indicates an error with the reflection. Generally this
     *             indicates a serious programming error and should not be
     *             caught
     */
    protected Object invokeWrappedServerMethod(RemoteServerMethod method,
    // CHECKSTYLE:OFF we really want to throw a runtime exception because this
            // indicates a serious programming error
            Object... args) throws InvocationTargetException, RuntimeException {
        // CHECKSTLYE:ON

        logger.debug("Calling " + method + " with arguments " + args);

        // if we did no get a remote server, we have to wait for it.
        if (getWrappedRemoteServer() == null) {
            try {
                waitForWrappedRemoteServerAsUserException();
            } catch (UserException e) {
                // for convenience reasons this is rethrown as an
                // InvocationTargetException so that every error handling
                // concerning the original server must only be done on the root
                // exception of the InvocationTargetException
                throw new InvocationTargetException(e);
            }
        }

        assert getWrappedRemoteServer() != null;
        Method m = getMethod(method);
        assert m != null : "After " + "the first remote server was set, "
                + "all methods must be known.";

        try {
            // hopefully this method will always throw an exception in
            // an error case. Otherwise we got a terrible dead lock.
            return m.invoke(getWrappedRemoteServer(), args);
        } catch (IllegalAccessException e) {
            logger.fatal(REFLECTION_ERROR_MESSAGE, e);
            assert false;
            // Rethrow this as a runtime exception to indicate a serious
            // programming error if this happens that should not be caught
            throw new RuntimeException(e);
        }

    }

    /**
     * Uses the {@link #reflectiveCallMethod(RemoteServerMethod, Object...)} but
     * swallows the {@link UserException} and {@link MethodNotFoundException}
     * exception.
     * 
     * @param method
     *            method to call
     * @param args
     *            method arguments
     * @return return value of method or <code>null</code> for void methods
     */
    protected Object callWithoutException(RemoteServerMethod method,
            Object... args) {
        try {
            return reflectiveCallMethod(method, args);
        } catch (UserException e) {
            logger.fatal("Must not be thrown here", e);
            assert false : "Must not be thrown here";
            return null;
        } catch (MethodNotFoundException e) {
            logger.fatal("Must not be thrown here", e);
            assert false : "Must not be thrown here";
            return null;
        }
    }

    /**
     * Returns the lock object that has to be used to work on the remote server
     * instance.
     * 
     * @return lock object
     */
    protected Object getRemoteServerLock() {
        return remoteServerLock;
    }

    /**
     * Returns the wrapped remote server.
     * 
     * @return wrapped remote server or <code>null</code> if there is no remote
     *         server yet
     */
    protected RemoteServer getWrappedRemoteServer() {
        synchronized (remoteServerLock) {
            return wrappedRemoteServer;
        }
    }

    /**
     * Marks the wrapped remote server as disconnected in the
     * {@link RemoteServerReconnectableManager}.
     */
    protected void markWrappedServerAsDisconnected() {
        manager.markRemoteServerAsDisconnected(wrappedRemoteServer
                .getServerName());
    }

    /**
     * Returns the {@link RemoteServerReconnectableManager} managing this
     * wrapper.
     * 
     * @return manager
     */
    protected RemoteServerReconnectableManager getManager() {
        return manager;
    }

    /**
     * Resets the remote server from a separate thread if the remote server was
     * disconnected.
     * 
     * @param remoteServer
     *            new remote server to wrap
     */
    void resetRemoteServer(RemoteServer remoteServer) {

        logger.info("Received new remote server to decorate: " + remoteServer);

        synchronized (getRemoteServerLock()) {

            wrappedRemoteServer = remoteServer;

            // analyze remote server for methods
            try {

                for (RemoteServerMethod methodKey : RemoteServerMethod.values()) {
                    methodsByUniqueName.put(methodKey, wrappedRemoteServer
                            .getClass().getMethod(methodKey.getMethodName(),
                                    methodKey.getSignature()));
                }

            } catch (SecurityException e) {
                logger.fatal(REFLECTION_ERROR_MESSAGE, e);
                assert false;
                throw new RuntimeException(REFLECTION_ERROR_MESSAGE, e);
            } catch (NoSuchMethodException e) {
                logger.fatal(REFLECTION_ERROR_MESSAGE, e);
                assert false;
                throw new RuntimeException(REFLECTION_ERROR_MESSAGE, e);
            } catch (NullPointerException e) {
                logger.fatal(REFLECTION_ERROR_MESSAGE, e);
                assert false;
                throw new RuntimeException(REFLECTION_ERROR_MESSAGE, e);
            }

            postProcessServerReset();

            logger.debug("Notifying waiting thread for new remote server.");
            remoteServerLock.notifyAll();

        }

    }

    /**
     * Hook method that can be implemented to do some actions after a server was
     * reset successfully.
     */
    protected void postProcessServerReset() {
        // nothing to do here
    }

    /**
     * Waits on {@link #remoteServerLock} for a new remote server. Always
     * acquire {@link #remoteServerLock} before using this method!
     * 
     * @throws InterruptedException
     *             interrupted while waiting
     */
    protected void waitForWrappedRemoteServer() throws InterruptedException {
        logger.debug("waiting for new remote server " + "to retry call again.");
        getRemoteServerLock().wait();
    }

    /**
     * Waits on {@link #remoteServerLock} for a new remote server and re-throws
     * the possible {@link InterruptedException} as {@link UserException}.
     * Always acquire {@link #remoteServerLock} before using this method!
     * 
     * @throws UserException
     *             interrupted while waiting
     */
    protected void waitForWrappedRemoteServerAsUserException()
            throws UserException {
        try {
            waitForWrappedRemoteServer();
        } catch (InterruptedException e) {
            throw new UserException(e);
        }
    }

    // ------------- original interface ----------------

    /**
     * {@inheritDoc}
     */
    @Override
    public void activate() throws XcfException {
        reflectiveCallMethod(RemoteServerMethod.ACTIVATE);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void addListener(SystemEventListener l) {
        callWithoutException(RemoteServerMethod.ADD_LISTENER, l);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void callMethod(String method, XOPData d, AsyncCallback cb)
            throws MethodNotFoundException, UserException {
        reflectiveCallMethod(RemoteServerMethod.CALL_METHOD_CALLBACK, method,
                d, cb);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void callMethod(String method, XOPData d, Xcf1Signature s,
            AsyncCallback cb) throws MethodNotFoundException, UserException {
        reflectiveCallMethod(RemoteServerMethod.CALL_METHOD_XCF1_CALLBACK,
                method, d, s, cb);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public XOPData callMethod(String method, XOPData d, Xcf1Signature s)
            throws MethodNotFoundException, UserException {
        return (XOPData) reflectiveCallMethod(
                RemoteServerMethod.CALL_METHOD_XCF1, method, d, s);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public XOPData callMethod(String method, XOPData d)
            throws MethodNotFoundException, UserException {
        return (XOPData) reflectiveCallMethod(
                RemoteServerMethod.NORMAL_CALL_METHOD, method, d);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void deactivate() {
        callWithoutException(RemoteServerMethod.DEACTIVATE);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(Object obj) {
        return this == obj;
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    @Override
    public Map getMethods() {
        return (Map) callWithoutException(RemoteServerMethod.GET_METHODS);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getServerName() {
        return (String) callWithoutException(RemoteServerMethod.GET_SERVER_NAME);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        return System.identityHashCode(this);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void removeListener(SystemEventListener l) {
        callWithoutException(RemoteServerMethod.REMOVE_LISTENER, l);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        synchronized (remoteServerLock) {
            String wrapped = "null";
            if (wrappedRemoteServer != null) {
                wrapped = wrappedRemoteServer.toString();
            }
            return getClass().getSimpleName() + ": " + wrapped;
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected ServerPrx getServer() {
        // XXX this is hopefully not used...
        return null;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void setServer(ServerPrx server) {
        // XXX this is hopefully not used...
    }

}
