package de.unibi.airobots.bonsai.xcf.actuators;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import net.sf.xcf.ActiveMemory;
import net.sf.xcf.RemoteServer;
import net.sf.xcf.XcfException;
import net.sf.xcf.transport.XOPData;
import net.sf.xcf.xml.XPath;
import nu.xom.Element;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.Globals;
import de.unibi.airobots.bonsai.actuators.HwaActuator;
import de.unibi.airobots.bonsai.xcf.ActiveMemoryReconnectable;
import de.unibi.airobots.bonsai.xcf.RemoteServerReconnectable;
import de.unibi.airobots.btl.data.PersonData;

/**
 * Actuator for controlling the HWA.
 * 
 * @author jwienke
 */
public class XcfHwaActuator implements HwaActuator, RemoteServerReconnectable,
        ActiveMemoryReconnectable {

    private static final String CONFIGURE_METHOD_NAME = "configure";

    private Logger logger = Logger.getLogger(getClass());

    private String remoteServerName = "";
    private String sceneMemoryName = "";
    private Object remoteServerLock = new Object();
    private Object sceneMemoryLock = new Object();
    private RemoteServer remoteServer = null;
    private ActiveMemory sceneMemory = null;

    /**
     * Constructor.
     * 
     * @param controlServerName
     *            desired name of the server for controlling the hwa.
     * @param sceneMemoryName
     *            desired name of the memory to control the hwa
     */
    public XcfHwaActuator(String controlServerName, String sceneMemoryName) {
        logger.debug("Creating HWAActuator.");
        this.remoteServerName = controlServerName;
        this.sceneMemoryName = sceneMemoryName;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void startCamAction() throws IOException {

        logger.debug("startCamAction called");

        try {
            XOPData result;
            synchronized (remoteServerLock) {
                result =
                        remoteServer.callMethod(CONFIGURE_METHOD_NAME, XOPData
                                .fromString("<MOVESRC cam=\"PTA\" "
                                        + "base=\"none\" />"));
            }

            logger.debug("Got result from HWA: " + result.getDocumentAsText());

            boolean good =
                    Boolean.parseBoolean(((Element) result.getDocument().query(
                            "//SUCCESS").get(0)).getAttributeValue("value"));
            if (!good) {
                throw new IOException(
                        "Could not start searching, remote server answer: "
                                + result.getDocumentAsText());
            }
        } catch (XcfException e) {
            throw new IOException(e);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void focus(PersonData p) throws IOException {

        try {
            unfocus();
            synchronized (sceneMemoryLock) {
                sceneMemory.insert(XOPData
                        .fromString("<IAP><IDENT name=\"\" id=\"" + p.getId()
                                + "\" /><GENERATOR>" + Globals.LIBRARY_NAME
                                + "</GENERATOR>"
                                + "<TIMESTAMP><INSERTED value=\""
                                + System.currentTimeMillis()
                                + "\"/><UPDATED value=\""
                                + System.currentTimeMillis()
                                + "\"/></TIMESTAMP></IAP>"));
            }

        } catch (XcfException e) {
            throw new IOException(e);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void unfocus() throws IOException {
        try {
            synchronized (sceneMemoryLock) {
                sceneMemory.remove(new XPath("//IAP"));
            }
        } catch (XcfException e) {
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void stopCamAction() throws IOException {

        logger.debug("stopCamAction called");

        try {

            XOPData result;
            synchronized (remoteServerLock) {
                result =
                        remoteServer.callMethod(CONFIGURE_METHOD_NAME, XOPData
                                .fromString("<MOVESRC cam=\"none\" "
                                        + "base=\"none\" />"));
            }

            logger.debug("Got result from HWA: " + result.getDocumentAsText());

            boolean good =
                    Boolean.parseBoolean(((Element) result.getDocument().query(
                            "//SUCCESS").get(0)).getAttributeValue("value"));
            if (!good) {
                throw new IOException(
                        "Could not stop searching, remote server answer: "
                                + result.getDocumentAsText());
            }
        } catch (XcfException e) {
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Map<String, ErrorHandlingPolicy> getRemoteServerNames() {
        Map<String, ErrorHandlingPolicy> servers =
                new HashMap<String, ErrorHandlingPolicy>();
        servers.put(remoteServerName, ErrorHandlingPolicy.SELF_HEALING);
        return servers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void resetRemoteServer(RemoteServer remoteServer) {
        assert remoteServer.getServerName().equals(remoteServer);
        synchronized (remoteServerLock) {
            this.remoteServer = remoteServer;
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getActiveMemoryName() {
        return sceneMemoryName;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void resetActiveMemory(ActiveMemory activeMemory) {
        assert activeMemory.getName().equals(sceneMemoryName);
        synchronized (sceneMemoryLock) {
            sceneMemory = activeMemory;
        }
    }

}
