package de.unibi.airobots.bonsai.xcf.configuration;

import java.util.Map;

import de.unibi.airobots.bonsai.core.Actuator;
import de.unibi.airobots.bonsai.core.CoreObjectFactory.ActuatorToConfigure;
import de.unibi.airobots.bonsai.core.CoreObjectFactory.CoreObjectCreationException;

/**
 * Base class for configuration objects that create actuators. For convenience
 * reasons the default implementation of the {@link #parseOptions(Map)} method
 * already parses a required server name.
 * 
 * @author jwienke
 */
public abstract class ConfiguredXcfActuator extends ConfiguredXcfCoreObject {

    /**
     * Option for remote server based classes defining the remote server name.
     */
    public static final String OPTION_SERVER_NAME = "serverName";

    /**
     * Name of the server if parsed by {@link #parseOptions(Map)}.
     */
    private String serverName = "";

    /**
     * Concrete class for the actuator requested by the user in the
     * configuration.
     */
    private Class<? extends Actuator> actuatorClass = null;

    /**
     * Override this!
     * 
     * @param actuatorToConfigure
     *            requested configuration
     * @return <code>true</code> if this configuration object is suitable for
     *         the requested actuator
     */
    public static boolean isSuitableFor(ActuatorToConfigure actuatorToConfigure) {
        return false;
    }

    /**
     * Create a new specific actuator instance.
     * 
     * @return new instance
     * @throws CoreObjectCreationException
     *             problem creating the new instance
     */
    public abstract Actuator createInstance()
            throws CoreObjectCreationException;

    /**
     * {@inheritDoc}
     * 
     * Default implementation parsing the server name.
     */
    @Override
    public void parseOptions(Map<String, String> options)
            throws IllegalArgumentException {

        if (!options.containsKey(OPTION_SERVER_NAME)) {
            throw new IllegalArgumentException(
                    "Server name missing in arguments for requested type");
        }

        this.serverName = options.get(OPTION_SERVER_NAME);

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String[] getRemoteServersToInitialize() {
        return new String[] { getServerName() };
    }

    /**
     * Returns the configured server name if this class'
     * {@link #parseOptions(Map)} method is used.
     * 
     * @return configured server name
     */
    public String getServerName() {
        return serverName;
    }

    /**
     * Setter for the implementation class requested by the user.
     * 
     * @param actuatorClass
     *            new actuator class
     */
    public void setActuatorClass(Class<? extends Actuator> actuatorClass) {
        this.actuatorClass = actuatorClass;
    }

    /**
     * Returns the concrete implementation class for the actuator requested in
     * the configuration.
     * 
     * @return implementation class for the actuator
     */
    public Class<? extends Actuator> getActuatorClass() {
        return actuatorClass;
    }

}
